/*
 * Copyright (c) 2010, Bruno Galerne <bruno.galerne@cmla.ens-cachan.fr>
 * All rights reserved.
 *
 * This program is free software: you can use, modify and/or
 * redistribute it under the terms of the GNU General Public
 * License as published by the Free Software Foundation, either
 * version 3 of the License, or (at your option) any later
 * version. You should have received a copy of this license along
 * this program. If not, see <http://www.gnu.org/licenses/>.
 */

/**
* @file random_phase_noise.c
* @brief Main function for random phase noise texture
* synthesis. Important modules are in the file @ref
* random_phase_noise_lib.c
*
* @version 1.3
* @author Bruno Galerne ; <bruno.galerne@cmla.ens-cachan.fr> ;
* <http://www.cmla.ens-cachan.fr/~galerne>
*/

/**
* @mainpage Source code for random phase noise texture synthesis
*
* @image html rpn_description.jpg "Random phase texture synthesis
* using ./random_phase_noise"
*
* This source code is an ANSI C implementation of the random phase
* noise texture synthesis algorithm describes in the IPOL webpage
* <a href="http://www.ipol.im/pub/algo/ggm_random_phase_texture_synthesis/"
* title="Go to the IPOL webpage Random Phase Texture Synthesis">Random
* Phase Texture Synthesis</a>.
*
* The last release should be available at
* <http://www.ipol.im/pub/algo/ggm_random_phase_texture_synthesis/>.
*
* This code is provided with an HTML documentation produced by doxygen
* <http://www.ipol.im/pub/algo/ggm_random_phase_texture_synthesis/srcdoc/>.
* 
* @author Bruno Galerne ; <bruno.galerne@cmla.ens-cachan.fr> ;
* <http://www.cmla.ens-cachan.fr/~galerne>
*
* @b Requirements: @n
* - ANSI C compiler
* - getopt
* - libpng
* - libfftw3
*
* @b Compilation: @n 
*
* Execute the provided Makefile (function make).
*
* @b Usage: (displayed in executing ./random_phase_noise -h)@n
*
* 	random_phase_noise -r ratio -x nxout -y nyout -a alpha -s seed
* 	image_in.png image_out.png @n
*
* 	<em> Required parameters:</em> @n
* 	image_in.png   :   name of the input PNG image @n
* 	image_out.png  :   name of the output PNG image @n
*		
*		<em> Optionnal parameters:</em> @n
*		ratio           :   float > 1 that specifies the increase
*                                   of size of image_out.png relatively to the
*		size of image_in.png @n 
*		nxout           :   int to specify the width of image_out.png.
*                                   Must be greater than the width of
*                                   image_in.png @n
*		nyout           :   int to specify the height of image_out.png.
*                                   Must be greater than the height of
*                                   image_in.png @n
*		alpha           :   float in (0,0.5) that specifies the
*                                   proportion of the smoothing zone
*                                   for the spot extension
*                                   (alpha = 0.1 by default) @n 
*		seed            :   unsigned int to specify the seed
*             		            for the random number generator
*                                   (seed = time(NULL) by default) @n 
*
* @b Test:
* To test the module run: @n
*    ./random_phase_noise -s 1 sample.png output.png @n
* output.png should be the same as the image sample_rpn.png provided
* with the source code.
*
* @b Releases:
*
* @li version 1.0:
* 	first version
*
* @li version 1.1:
* 	(optimization) Use the FFTW functions fftw_plan_dft_r2c_2d and
* 	fftw_plan_dft_c2r_2d for discrete Fourier transforms
* 	computation (reduce the size of all complex arrays).
*
* @li version 1.2:
*		New version of the spot extension technique (see the
*		@ref spot_extension_color function):
* 	   For the smooth transition the cumulative histogram of a
*		standard molifier is used instead of a Gaussian
*		decay. 
* 	   This is mathematically more coherent, but does not lead to
*		significant changes for the produced textures.
*    In addition, the new option -s enables to specify the seed for
*		the random number generator.
*		This permits to use the same random phase with
*		different input images.
*    
* @li version 1.21:
*    The source files have been reorganized: all the relevant
*    functions are now in the file @ref random_phase_noise_lib.c.
*    Besides, the source code for png input/output is a more recent
*    version due to Nicolas Limare
*    <nicolas.limare@cmla.ens-cachan.fr>.
*
* @li version 1.3:
*    Minor bug fixed thanks to the reviewer (unsigned int different
*    from size_t for 64 bits).
*    A small png image and its corresponding output are added to the
*    release to enable a quick test of the module.
*    Change of input/output routines: input and output images have now
*    the .png image format.
*
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <getopt.h>
#include <fftw3.h>

#include "mt.h"
#include "io_png.h"
#include "random_phase_noise_lib.h"

/** print a message and abort the execution */
#define FATAL(MSG)			\
do {				\
	fprintf(stderr, MSG "\n");	\
	abort();			\
	} while (0);

/*
* Main function section
*/

/**
* @brief Display the usage of the main module on stdout.
*/

static int display_usage()
{
    printf("\nrandom_phase_noise -r ratio -x nxout -y nyout "
           "-a alpha -s seed image_in.png image_out.png\n\n");
    printf("Required parameters:\n");
    printf("   image_in.png   :   name of the input PNG image\n");
    printf("   image_out.png  :   name of the output PNG image\n\n");
    printf("Optionnal parameters:\n");
    printf("   ratio           :   float > 1 that specifies the increase "
           "of size of image_out.png relatively to the size "
           "of image_in.png\n");
    printf("   nxout           :   int to specify the width "
           "of image_out.png. Must be greater than the width "
           "of image_in.png\n");
    printf("   nyout           :   int to specify the height "
           "of image_out.png. Must be greater than the height "
           "of image_in.png\n");
    printf("   alpha           :   float in (0,0.5) that specifies "
           "the proportion of the smoothing zone for "
           "the spot extension (alpha = 0.1 by default)\n");
    printf("   seed            :   unsigned int to specify the seed "
           "for the random number generator "
           "(seed = time(NULL) by default)\n");

    return (1);
}

/**
* @brief main function call
*/
int main(int argc, char **argv)
{
    char *fname_in, *fname_out; /* input/output file names */
    float *data_in;             /*input data */
    float *data_in_rgb[3];
    float *data_out;            /* output data */
    float *data_out_rgb[3];
    size_t nxin, nyin;          /* input size */
    size_t nxout, nyout;        /* output size */
    float alpha;
    int r_flag = 0;
    float ratio;                /* increase ratio for the
                                 * output size (must be >1) */
    unsigned long seed = 0;     /* seed for the random number
                                 * generator */
    int x_flag = 0;
    int y_flag = 0;
    int a_flag = 0;
    int s_flag = 0;
    int c;                      /* getopt flag */

    /* "default" value initialization */
    alpha = 0.1;
    nxout = 0;
    nyout = 0;
    ratio = 1.;

    /* process the options and parameters */

    while ((c = getopt(argc, argv, "r:x:y:a:s:hv")) != -1) {
        switch (c) {
        case 'r':
            /* increase ratio specified */
            r_flag = 1;
            ratio = atof(optarg);
            if (ratio <= 1)
                FATAL("Option usage: The ratio "
                      "specified by -r must be " "greater than 1.")
                    break;
        case 'x':
            /* output width specified */
            x_flag = 1;
            nxout = atoi(optarg);
            break;
        case 'y':
            /* output height specified */
            y_flag = 1;
            nyout = atoi(optarg);
            break;
        case 'a':
            /* alpha parameter specified */
            a_flag = 1;
            alpha = atof(optarg);
            break;
        case 's':
            /* seed specified */
            s_flag = 1;
            seed = (unsigned long) atol(optarg);
            break;
        case 'h':
            /* display usage */
            display_usage();
            return (-1);
        case 'v':
            /* display version */
            fprintf(stdout, "%s version " __DATE__ "\n", argv[0]);
            return (-1);
        default:
            abort();
        }
    }

    /* process the non-option parameters */
    if (2 > (argc - optind)) {
        printf("The image file names are missing\n\n");
        display_usage();
        return (-1);
    }
    fname_in = argv[optind++];
    fname_out = argv[optind++];

    /* initilization of the random number generator */
    if (s_flag == 0) {
        mt_init((unsigned long) time(NULL));
    }
    else {
        mt_init(seed);
    }

    /* read the PNG image and set pointers for data_in */
    if (NULL == (data_in = io_png_read_f32_rgb(fname_in, &nxin, &nyin)))
        FATAL("error while reading the PNG image");
    data_in_rgb[0] = data_in;
    data_in_rgb[1] = data_in + nxin * nyin;
    data_in_rgb[2] = data_in + 2 * nxin * nyin;

    /* deal with options conflict/inconsistance and set the output size */
    if ((r_flag == 1) && ((x_flag == 1) || (y_flag == 1))) {
        printf("Option usage: Conflicting options to specify "
               "the output size: Use either the -r option "
               "or both -x and -y options\n\n");
        display_usage();
        return (-1);
    }
    if (x_flag == 0) {
        nxout = (int) (ratio * ((float) nxin));
    }
    else if (nxin > nxout) {
        printf("Option usage: the width specified by "
               "the -x option must be greater than the one "
               "of the input image\n\n");
        display_usage();
        return (-1);
    }
    if (y_flag == 0) {
        nyout = (int) (ratio * ((float) nyin));
    }
    else if (nyin > nyout) {
        printf("Option usage: the height specified by "
               "the -y option must be greater than the one "
               "of the input image\n\n");
        display_usage();
        return (-1);
    }

    if ((a_flag == 1) && (nxin == nxout) && (nyin == nyout)) {
        printf("INFO: The -a option was ignored "
               "since the output size is the same as " "the input size.\n");
    }

    /* Two modes depending whether the input size and the output
     * size are the same */
    if ((nxin == nxout) && (nyin == nyout)) {   /* Same size */
        /* RPN computation */
        if (NULL == rpn_color_same_size(data_in_rgb, nxin, nyin))
            FATAL("error while running rpn_color_same_size");

        /* Write result in PNG file */
        if (0 != io_png_write_f32(fname_out, data_in, nxin, nyin, 3))
            FATAL("error while writing output PNG file");

        /* Free memory */
        free(data_in);
    }
    else {                      /* Different sizes */

        /* Memory allocation for data_out */
        if (NULL == (data_out = (float *)
                     malloc(3 * nxout * nyout * sizeof(float))))
            FATAL("allocation error");
        data_out_rgb[0] = data_out;
        data_out_rgb[1] = data_out + nxout * nyout;
        data_out_rgb[2] = data_out + 2 * nxout * nyout;

        /* RPN computation */
        if (NULL == rpn_color_different_size(data_out_rgb, data_in_rgb,
                                             nxout, nyout, nxin, nyin, alpha))
            FATAL("error while running rpn_color_different_size");

        /* Write result in PNG file */
        if (0 != io_png_write_f32(fname_out, data_out, nxout, nyout, 3))
            FATAL("error while writing output PNG file");

        /* Free memory */
        free(data_in);
        free(data_out);
    }

    return (0);
}
